#--
#-- SUB: createRemoteSnapshot
#-- Creates a snapshot based on a remote database query
#--
sub createRemoteSnapshot {
	#-- Function parameters
	my ($dbh, $schemaname, $snapshotname, $query, $dblinkname) = @_;

	#-- Local variables
	my $sql;
	my $sth;
	my $dblink;
	my $attr_href;

	#--Test if the DBLINK entry exists
	$sql = 'SELECT * FROM %BASE_SCHEMA%.pg_dblinks WHERE dblinkname=?';
	$sth = $dbh->prepare($sql);
	$sth->execute(($dblinkname));
	if ($sth->err) {
		elog ERROR, "Table '%BASE_SCHEMA%.pg_dblinks' does not exist!";
		return FALSE;
	}

	#-- Fetch the dblink info
	$dblink = $sth->fetchrow_hashref;
	if (! $dblink->{'dblinkname'}) {
		elog ERROR, "DBLink '$dblinkname' does not exist!";
		return FALSE;
	}

	#-- Convert the attributes to a hashtable
	$attr_href = eval($dblink->{'attributes'});

	#-- Make the remote database connection
	$dbh = dbiGetConnection(
		$dblink->{'datasource'}
		, $dblink->{'username'}
		, $dblink->{'password'}
		, $attr_href
		);

	#-- holds each column definition of the new table
	my @cols;

	#-- retrieve the structure of the snapshot query (empty query)
	#-- some fake databases do not support subqueries, so we need to modify
	#-- the original query and insert "1=0 AND " or "WHERE 1=0" as needed
	#-- of course, we try "SELECT * FROM ($query) WHERE 1=0" first,
	#-- for databases that support subqueries, or better: REAL databases
	$sql = 'SELECT * FROM ($query) query WHERE 1=0';
	elog NOTICE, $sql if $main::DEBUG == 1;
	$sth = $dbh->prepare($sql);
	if ($DBI::errstr) {
		$query =~ tr/\n/ /;
		my $lc_query = lc($query);
		my $wherepos = rindex($lc_query, " where ");
		if ($wherepos > 0) {
			$sql = substr($query, 0, $wherepos) . ' WHERE 1=0 AND ' . substr($query, $wherepos + length(" where "));
		} else {
			my $orderbypos = rindex($lc_query, " order by ");
			my $groupbypos = rindex($lc_query, " group by ");
			if ($orderbypos < $groupbypos) {
				$wherepos = $orderbypos;
			} else {
				$wherepos = $groupbypos;
			}
			if ($wherepos == -1) {
				$wherepos = length($lc_query);
			}
			$sql = substr($query, 0, $wherepos) . ' WHERE 1=0 ' . substr($query, $wherepos);
		}
	
		elog NOTICE, $sql if $main::DEBUG == 1;
		$sth = $dbh->prepare($sql);
		if ($DBI::errstr) {
			my $err = <<ERR2;
Cannot prepare

$sql

$DBI::errstr
ERR2
			elog ERROR, $err;
		}
	}

	$sth->execute;

	#--------------------------------
	#-- SQL datatype codes
	#-- SQL_GUID                         (-11)
	#-- SQL_WLONGVARCHAR                 (-10)
	#-- SQL_WVARCHAR                      (-9)
	#-- SQL_WCHAR                         (-8)
	#-- SQL_BIT                           (-7)
	#-- SQL_TINYINT                       (-6)
	#-- SQL_BIGINT                        (-5)
	#-- SQL_LONGVARBINARY                 (-4)
	#-- SQL_VARBINARY                     (-3)
	#-- SQL_BINARY                        (-2)
	#-- SQL_LONGVARCHAR                   (-1)
	#-- SQL_UNKNOWN_TYPE                    0
	#-- SQL_ALL_TYPES                       0
	#-- SQL_CHAR                            1
	#-- SQL_NUMERIC                         2
	#-- SQL_DECIMAL                         3
	#-- SQL_INTEGER                         4
	#-- SQL_SMALLINT                        5
	#-- SQL_FLOAT                           6
	#-- SQL_REAL                            7
	#-- SQL_DOUBLE                          8
	#-- SQL_DATETIME                        9
	#-- SQL_DATE                            9
	#-- SQL_INTERVAL                       10
	#-- SQL_TIME                           10
	#-- SQL_TIMESTAMP                      11
	#-- SQL_VARCHAR                        12
	#-- SQL_BOOLEAN                        16
	#-- SQL_UDT                            17
	#-- SQL_UDT_LOCATOR                    18
	#-- SQL_ROW                            19
	#-- SQL_REF                            20
	#-- SQL_BLOB                           30
	#-- SQL_BLOB_LOCATOR                   31
	#-- SQL_CLOB                           40
	#-- SQL_CLOB_LOCATOR                   41
	#-- SQL_ARRAY                          50
	#-- SQL_ARRAY_LOCATOR                  51
	#-- SQL_MULTISET                       55
	#-- SQL_MULTISET_LOCATOR               56
	#-- SQL_TYPE_DATE                      91
	#-- SQL_TYPE_TIME                      92
	#-- SQL_TYPE_TIMESTAMP                 93
	#-- SQL_TYPE_TIME_WITH_TIMEZONE        94
	#-- SQL_TYPE_TIMESTAMP_WITH_TIMEZONE   95
	#-- SQL_INTERVAL_YEAR                 101
	#-- SQL_INTERVAL_MONTH                102
	#-- SQL_INTERVAL_DAY                  103
	#-- SQL_INTERVAL_HOUR                 104
	#-- SQL_INTERVAL_MINUTE               105
	#-- SQL_INTERVAL_SECOND               106
	#-- SQL_INTERVAL_YEAR_TO_MONTH        107
	#-- SQL_INTERVAL_DAY_TO_HOUR          108
	#-- SQL_INTERVAL_DAY_TO_MINUTE        109
	#-- SQL_INTERVAL_DAY_TO_SECOND        110
	#-- SQL_INTERVAL_HOUR_TO_MINUTE       111
	#-- SQL_INTERVAL_HOUR_TO_SECOND       112
	#-- SQL_INTERVAL_MINUTE_TO_SECOND     113
	#--------------------------------

	#-- Loop through all columns and retrieves the name and type of each one
	#-- Also maps the SQL Perl types to PostgreSQL ones
	#-- Place this info on @cols
	for ( my $i = 0 ; $i < $sth->{NUM_OF_FIELDS} ; $i++ ) {
            my $line = $sth->{NAME}->[$i];
	    my $datatype = $sth->{TYPE}->[$i];
	    my $precision = $sth->{PRECISION}->[$i];
	    my $scale = $sth->{SCALE}->[$i];
	    my $isnullable = $sth->{NULLABLE}->[$i];
	    my $nullable = '';
	    if (! $isnullable) {
		$nullable = 'NOT NULL';
	    }

            if ($datatype == DBI::SQL_INTEGER ) {
		$line .= " INTEGER $nullable";
	    } elsif ($datatype == DBI::SQL_SMALLINT ) {
		$line .= " SMALLINT $nullable";
	    } elsif (($datatype == DBI::SQL_NUMERIC )
		|| ($datatype == DBI::SQL_DECIMAL )
		|| ($datatype == DBI::SQL_FLOAT )
		|| ($datatype == DBI::SQL_REAL )
		|| ($datatype == DBI::SQL_DOUBLE )
#--		|| ($datatype == DBI::SQL_BIGINT )
		|| ($datatype == DBI::SQL_TINYINT )) { #--NUMBER|NUMBER(?)
                $line .= " NUMERIC($precision, $scale) $nullable";
	    } elsif ( ( $datatype == DBI::SQL_BOOLEAN ) 
			|| ( $datatype == DBI::SQL_BIT ) ) { #--BOOLEAN
		$line .= " BOOLEAN $nullable";
            } elsif ( ($datatype == DBI::SQL_CHAR )
			|| ( $datatype == DBI::SQL_WCHAR ) ) { #--CHAR
		$line .= " CHAR($precision) $nullable";
	    } elsif ( ($datatype == DBI::SQL_VARCHAR )
			|| ( $datatype == DBI::SQL_WVARCHAR ) ) {#--VARCHAR
		$line .= " VARCHAR($precision) $nullable";
	    } elsif ( ( $datatype == DBI::SQL_LONGVARCHAR )
			|| ( $datatype == DBI::SQL_WLONGVARCHAR )
			|| ( $datatype == DBI::SQL_CLOB ) ) { #--CLOB
		$line .= " VARCHAR $nullable";
	    } elsif ( ( $datatype == DBI::SQL_DATE )
			|| ( $datatype == DBI::SQL_TYPE_DATE ) ) { #--DATE
		$line .= " DATE $nullable";
	    } elsif ( ( $datatype == DBI::SQL_TIME )
			|| ( $datatype == DBI::SQL_TYPE_TIME ) ) { #--TIME
		$line .= " TIME $nullable";
	    } elsif ( ( $datatype == DBI::SQL_TYPE_TIME_WITH_TIMEZONE ) ) {
		$line .= " TIME WITH TIMEZONE $nullable";
	    } elsif ( ( $datatype == DBI::SQL_DATETIME )
			|| ( $datatype == DBI::SQL_TIMESTAMP )
			|| ( $datatype == DBI::SQL_TYPE_TIMESTAMP ) ) { #--TIMESTAMP
		$line .= " TIMESTAMP $nullable";
	    } elsif ( ( $datatype == DBI::SQL_TYPE_TIMESTAMP_WITH_TIMEZONE ) ) {
		$line .= " TIMESTAMP WITH TIMEZONE $nullable";
	    } elsif ( ( $datatype == DBI::SQL_INTERVAL )
			|| ( $datatype == DBI::SQL_INTERVAL_YEAR )
			|| ( $datatype == DBI::SQL_INTERVAL_MONTH )
			|| ( $datatype == DBI::SQL_INTERVAL_DAY )
			|| ( $datatype == DBI::SQL_INTERVAL_HOUR )
			|| ( $datatype == DBI::SQL_INTERVAL_MINUTE )
			|| ( $datatype == DBI::SQL_INTERVAL_SECOND )
			|| ( $datatype == DBI::SQL_INTERVAL_YEAR_TO_MONTH )
			|| ( $datatype == DBI::SQL_INTERVAL_DAY_TO_HOUR )
			|| ( $datatype == DBI::SQL_INTERVAL_DAY_TO_MINUTE )
			|| ( $datatype == DBI::SQL_INTERVAL_DAY_TO_SECOND )
			|| ( $datatype == DBI::SQL_INTERVAL_HOUR_TO_MINUTE )
			|| ( $datatype == DBI::SQL_INTERVAL_HOUR_TO_SECOND )
			|| ( $datatype == DBI::SQL_INTERVAL_MINUTE_TO_SECOND ) ) {
		$line .= " INTERVAL $nullable";
            } elsif ( ( $datatype == DBI::SQL_BINARY )
			|| ( $datatype == DBI::SQL_VARBINARY )
			|| ( $datatype == DBI::SQL_LONGVARBINARY )
			|| ( $datatype == DBI::SQL_BLOB ) ) { #--BLOB
		$line .= " BYTEA $nullable";
	    } else { #--UNKNOWN
		elog WARNING, "Unknown type '$datatype' mapped to 'TEXT'!";
		$line .= " TEXT $nullable";
            }
            push @cols, $line;
	}
	$sth->finish;

	#-- Makes the CREATE TABLE SQL statement
        $sql = <<SQL;
CREATE TABLE $schemaname.$snapshotname (
  @{[ join("\n, ", @cols) ]}
)
SQL
	elog NOTICE, $sql if $main::DEBUG==1;
	$sth = spi_exec_query($sql);
	if ($sth->{status} eq 'SPI_OK_UTILITY') {
		elog NOTICE, "Snapshot placeholder created" if $main::DEBUG==1;
	} else {
		elog ERROR, "Could not create snapshot placeholder: $sql [$sth->{status}]";
	}
	#-- Disconnects from the remote database
	$dbh->disconnect;
}
